@@ -1,3 +1,4 @@
+#define PY_SSIZE_T_CLEAN
 #include <Python.h>
 #include <datetime.h>
 #include <iostream>
@@ -13,6 +14,32 @@
 #define TIME_LEN 10
 #define RANDOM_LEN 16
 
+#if PY_VERSION_HEX < 0x03070000
+PyObject* timezone_utc;
+#define UTC_SINGLETON timezone_utc
+void create_utc_tz() {
+  PyObject* pytz = PyImport_ImportModule("datetime");
+  if (!pytz) {
+    return;
+  }
+  PyObject* pytz_utc = PyObject_GetAttrString(pytz, "timezone");
+  Py_DECREF(pytz);
+  if (!pytz_utc) {
+    return;
+  }
+  PyObject* pytz_utc_utc = PyObject_GetAttrString(pytz_utc, "utc");
+  Py_DECREF(pytz_utc);
+  timezone_utc = pytz_utc_utc;
+  Py_DECREF(pytz_utc_utc);
+}
+#define CALLER_HEAD(f) (PyCFunction)f
+#define METH_MAYBE_FASTCALL METH_VARARGS
+#else
+#define UTC_SINGLETON PyDateTime_TimeZone_UTC
+#define CALLER_HEAD(f) (PyCFunction)(void(*)(void))f
+#define METH_MAYBE_FASTCALL METH_FASTCALL
+#endif
+
 std::random_device rnd;
 unsigned long long lastTime = 0;
 uint8_t *lastRandom = nullptr;
@@ -158,142 +185,22 @@ double decodeTime(const char* ulid) {
 }
 
 
-static PyObject *ulid_py(PyObject *self, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
-    static char *keywords[] = {"timestamp", NULL};
-
-    PyObject *arg = nullptr;
-    if (nargs == 1){
-        arg = args[0];
-    } else if (kwnames) {
-        if (PyTuple_Size(kwnames) == 1) {
-            PyObject *pyArgName = PyTuple_GetItem(kwnames, 0);
-            auto argName = PyUnicode_AsUTF8(pyArgName);
-            if (strcmp(keywords[0], argName) != 0 ) {
-                PyErr_SetString(PyExc_TypeError, "Invalid keyword argument for ulid()");
-                return NULL;
-            }
-            arg = args[0];
-        } else if (PyTuple_Size(kwnames) > 1) {
-            PyErr_SetString(PyExc_TypeError, "ulid() takes at most 1 keyword argument");
-            return NULL;
-        }
-    }
-
-    if (arg == nullptr) {
-        auto result = ULID(NULL);
-        if (result == nullptr) {
-            return NULL;
-        }
-        return Py_BuildValue("s", result);
-    }
-
-    double timestamp;
-    if (PyDateTime_Check(arg)) {
-        PyObject *py_timestamp = PyObject_CallMethod(arg, "timestamp", NULL);
-        timestamp = PyFloat_AsDouble(py_timestamp);
-        Py_DECREF(py_timestamp);
-    } else if (PyFloat_Check(arg)) {
-        timestamp = PyFloat_AsDouble(arg);
-    } else {
-        PyErr_SetString(PyExc_TypeError, "Argument must be a instance of `datetime.datetime` or `float`");
-        return NULL;
-    }
-
-    auto result = ULID(timestamp * 1000);
-    if (result == nullptr) {
-        return NULL;
-    }
-    return Py_BuildValue("s", result);
-}
-
-static PyObject *decode_datetime_py(PyObject *self, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
-    static char *keywords[] = {"ulid", NULL};
-
-    const char *ulid;
-
-    PyObject *arg = nullptr;
-    if (nargs == 1){
-        arg = args[0];
-    } else if (kwnames) {
-        if (PyTuple_Size(kwnames) == 1) {
-            PyObject *pyArgName = PyTuple_GetItem(kwnames, 0);
-            auto argName = PyUnicode_AsUTF8(pyArgName);
-            if (strcmp(keywords[0], argName) != 0 ) {
-                PyErr_SetString(PyExc_TypeError, "Invalid keyword argument for decode_datetime()");
-                return NULL;
-            }
-            PyArg_ParseTuple(*args, "s", &ulid);
-        } else if (PyTuple_Size(kwnames) > 1) {
-            PyErr_SetString(PyExc_TypeError, "decode_datetime() takes at most 1 keyword argument");
-            return NULL;
-        }
-    }
-
-    if (!PyUnicode_AsUTF8(arg)){
-        PyErr_SetString(PyExc_TypeError, "Argument must be a instance of `str`");
-        return NULL;
-    }
-    ulid = PyUnicode_AsUTF8(arg);
-
-    try {
-        double timestamp = decodeTime(ulid);
-        if (timestamp < 0) return NULL;
-        auto datetime = PyDateTime_FromTimestamp(Py_BuildValue("(d, O)", timestamp, PyDateTime_TimeZone_UTC));
-        if (datetime == NULL) {
-            PyErr_SetString(PyExc_OverflowError, "Timestamp is too large to create `datetime.datetime`. Please use `decode_timestamp` instead.");
-        }
-        return datetime;
-    } catch (std::runtime_error &e) {
-        PyErr_SetString(PyExc_RuntimeError, e.what());
-    }
-}
-
-static PyObject *decode_timestamp_py(PyObject *self, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
-    static char *keywords[] = {"ulid", NULL};
-
-    const char *ulid;
-
-    PyObject *arg = nullptr;
-    if (nargs == 1){
-        arg = args[0];
-    } else if (kwnames) {
-        if (PyTuple_Size(kwnames) == 1) {
-            PyObject *pyArgName = PyTuple_GetItem(kwnames, 0);
-            auto argName = PyUnicode_AsUTF8(pyArgName);
-            if (strcmp(keywords[0], argName) != 0 ) {
-                PyErr_SetString(PyExc_TypeError, "Invalid keyword argument for decode_datetime()");
-                return NULL;
-            }
-            PyArg_ParseTuple(*args, "s", &ulid);
-        } else if (PyTuple_Size(kwnames) > 1) {
-            PyErr_SetString(PyExc_TypeError, "decode_datetime() takes at most 1 keyword argument");
-            return NULL;
-        }
-    }
-
-    if (!PyUnicode_AsUTF8(arg)){
-        PyErr_SetString(PyExc_TypeError, "Argument must be a instance of `str`");
-        return NULL;
-    }
-    ulid = PyUnicode_AsUTF8(arg);
-
-    try {
-        double timestamp = decodeTime(ulid);
-        if (timestamp < 0) return NULL;
-        return Py_BuildValue("d", timestamp);
-    } catch (std::runtime_error &e) {
-        PyErr_SetString(PyExc_RuntimeError, e.what());
-    }
-}
-
 static char ulid_docs[] = "ulid(timestamp: datetime.datetime | float = None)-> str: \nReturn ULID. if timestamp is given, create ULID with it.\n";
 static char decode_datetime_docs[] = "decode_datetime(ulid: str)-> datetime.datetime: \nReturn timestamp of ULID in `datetime.datetime`.\nTimestamp can cause overflow, so if it's too large, please use `decode_timestamp` instead.\n";
 static char decode_timestamp_docs[] = "decode_timestamp(ulid: str)-> float: \nReturn timestamp of ULID in `float`.\n";
 
+
+#if PY_VERSION_HEX < 0x03070000
+#include "ULID_below_36.cpp"
+#else
+#include "ULID_above_37.cpp"
+#endif
+
+
 static PyMethodDef ulid_module_methods[] = {
-        {"ulid",        (PyCFunction)(void(*)(void))ulid_py,        METH_FASTCALL | METH_KEYWORDS, ulid_docs},
-        {"decode_datetime", (PyCFunction)(void(*)(void))decode_datetime_py, METH_FASTCALL | METH_KEYWORDS, decode_datetime_docs},
-        {"decode_timestamp", (PyCFunction)(void(*)(void))decode_timestamp_py, METH_FASTCALL | METH_KEYWORDS, decode_timestamp_docs},
+        {"ulid",        CALLER_HEAD(ulid_py),        METH_MAYBE_FASTCALL | METH_KEYWORDS, ulid_docs},
+        {"decode_datetime", CALLER_HEAD(decode_datetime_py), METH_MAYBE_FASTCALL | METH_KEYWORDS, decode_datetime_docs},
+        {"decode_timestamp", CALLER_HEAD(decode_timestamp_py), METH_MAYBE_FASTCALL | METH_KEYWORDS, decode_timestamp_docs},
         {NULL,          NULL,                         0,                            NULL}
 };
 
@@ -308,5 +215,8 @@ static struct PyModuleDef ulid_module_definition = {
 PyMODINIT_FUNC PyInit_fast_ulid(void) {
     Py_Initialize();
     PyDateTime_IMPORT;
+    #if PY_VERSION_HEX < 0x03070000
+    create_utc_tz();
+    #endif
     return PyModule_Create(&ulid_module_definition);
 }
\ No newline at end of file
